--[[
    Diddy Kong Racing Self-Magnet Item by yoshimo
    For X-Item Lib by minenice
    
    v0.2.1:
        Fixed forgetting to change a function for new implementation
    
    v0.2:
        Fixed places where checking for player being alive didn't check if player is not respawning
        Same as above but the inverse
        Fixed attempting to iterate over magnet info without checking if nil
        Fixed magnet info not updating while conditions to stop targeting were met
    
    v0.1.1:
        Fixed player variable initialization for players that join after add-on load during a match
    
    v0.1:
        Initial
]]

freeslot(
    "MT_DKRRETICLEFRAME",
    "MT_DKRRETICLEGLASS",
    "MT_DKRMAGNETELECTRICITY",
    "sfx_atdrmg"
)

--cache enumerated global constants
--commented vars are only ever referenced once and are for future reference
--local ANG15 = ANG15
--local ANGLE_315 = ANGLE_315
local FRACBITS = FRACBITS
local MT_PLAYER = MT_PLAYER
local TICRATE = TICRATE
local k_invincibilitytimer = k_invincibilitytimer
local k_growshrinktimer = k_growshrinktimer
local PST_LIVE = PST_LIVE
local k_hyudorotimer= k_hyudorotimer
local k_respawn = k_respawn
local MT_DKRRETICLEFRAME = MT_DKRRETICLEFRAME
local MT_DKRRETICLEGLASS = MT_DKRRETICLEGLASS
local MT_DKRMAGNETELECTRICITY = MT_DKRMAGNETELECTRICITY
local k_itemamount = k_itemamount
local ANGLE_90 = ANGLE_90
local MFE_VERTICALFLIP = MFE_VERTICALFLIP
local sfx_atdrmg = sfx_atdrmg
--local SPR_ADSM = SPR_ADSM

--must be [180, 360) to account for 180 degrees always being negative in Lua
local TARGETING_FORCE_ANGLE_THRES = -ANG15
local TARGETING_ANY_ANGLE_THRES = ANGLE_315

local TARGETING_DIST_THRES = 1024 << FRACBITS
local MAGNET_KILL_DIST_THRES = mobjinfo[MT_PLAYER].radius << 3
local MAGNET_TARGET_OFFSET_DIST = mobjinfo[MT_PLAYER].radius << 2

local MAGNET_PULL_FORCE = 40 << FRACBITS
local MAGNET_TIME = TICRATE << 1
local MAGNET_ELECTRICITY_COUNT = 8
local MAGNET_ELECTRICITY_OFFSET = 8 << FRACBITS
local MAGNET_SOUND_LOOP = TICRATE * 712 / 2205 --approximation, truncated

--mapobjectscaled vars
local targetingDistThresScaled
local magnetKillDistThresScaled
local magnetTargetOffsetDistScaled
local magnetPullForceScaled

local defrostingDebounce = false

local function netSyncLocals(network)
    targetingDistThresScaled = network(targetingDistThresScaled)
    magnetKillDistThresScaled = network(magnetKillDistThresScaled)
    magnetPullForceScaled = network(magnetPullForceScaled)
    magnetTargetOffsetDistScaled = network(magnetTargetOffsetDistScaled)
    defrostingDebounce = network(defrostingDebounce)
end
addHook("NetVars", netSyncLocals)

local function getMapObjectScaled(a)
    return FixedMul(a, mapobjectscale)
end

local function ternaryOperation(c, t, f)
    local r = f
    if c then
        r = t
    end
    return r
end

local function randomInt()
    return (P_RandomFixed() << FRACBITS) | P_RandomFixed()
end

--thanks call
--inserted here by mine
local function generateHFForPlayer(p)
    if p and p.valid and p.mo and p.mo.valid then
        return {{p.name, p.skincolor}, {skins[p.mo.skin].facemmap, p.skincolor}, p}
    else
        return {{}, {}}
    end
end

local function canDKRBackMagnetPlayer(player)
    return player.kartstuff[k_invincibilitytimer] <= 0
        and player.kartstuff[k_growshrinktimer] <= 0
end

local function canDKRTargetPlayer(player, otherP)
    return player ~= otherP
        and otherP
        and otherP.valid
        and otherP.mo
        and otherP.mo.valid
        and not otherP.spectator
        and otherP.playerstate == PST_LIVE
        and not otherP.kartstuff[k_hyudorotimer]
        and not otherP.kartstuff[k_respawn]
        and canDKRBackMagnetPlayer(otherP)
end

local function isMagnetAtInvalidRange(player, target)
    return R_PointToDist2(player.mo.x, player.mo.y, target.mo.x, target.mo.y)
            <= player.mo.radius + magnetKillDistThresScaled + target.mo.radius
        --[[or -abs(R_PointToAngle2(player.mo.x, player.mo.y, target.mo.x, target.mo.y))
            < ANGLE_270 + ANG20]]
end

local function mustKillAllMagnets(player)
    return
        not (player.mo and player.mo.valid) 
        or player.spectator
        or player.playerstate ~= PST_LIVE
        or player.kartstuff[k_respawn] ~= 0
end

local function isMagnetTargetValid(target)
    return
        not (target and target.valid and target.mo and target.mo.valid)
        or target.spectator
        or target.playerstate ~= PST_LIVE
        or target.kartstuff[k_respawn] ~= 0
end

local function relocateTargetPiece(piece, x, y, z, oldTarget)
    --reposition reticle
    P_MoveOrigin(piece, x, y, z)
    
    --INTERP: trick it into moving with old target if swapping targets
    if target == oldTarget then
        --makes it default to interp to new position
        piece.momx = 0
        piece.momy = 0
        piece.momz = 0
    elseif
        oldTarget
        and oldTarget.valid
        and oldTarget.mo
        and oldTarget.mo.valid
        and not oldTarget.spectator
    then
        local oldTargetMo = oldTarget.mo
        local x = oldTargetMo.momx
        local y = oldTargetMo.momy
        local z = oldTargetMo.momz
        if x or y or z then
            piece.momx = x
            piece.momy = y
            piece.momz = z
        else
            --trick it into barely moving at all for this tic
            piece.momz = 1
        end
    else
        --trick it into barely moving at all for this tic
        piece.momz = 1
    end
end

local function refreshDKRTargetStuff(player)
    local playerMo = player.mo
    local target
    local closestAngle = TARGETING_FORCE_ANGLE_THRES - 1
    local closestDist = targetingDistThresScaled + 1
    
    for i = 0, #players - 1 do
        local otherP = players[i]
        
        if canDKRTargetPlayer(player, otherP) then
            local dist = R_PointToDist2(playerMo.x, playerMo.y, otherP.mo.x, otherP.mo.y)
            
            if dist <= targetingDistThresScaled then
                local angle = -abs(R_PointToAngle2(
                    playerMo.x,
                    playerMo.y,
                    otherP.mo.x,
                    otherP.mo.y
                ) - playerMo.angle)
                
                if angle > closestAngle then
                    target = otherP
                    closestAngle = angle
                    --since someone is within the "scope" range, we only prioritize angle now
                    for j = i + 1, #players - 1 do
                        otherP = players[i]
                        if canDKRTargetPlayer(player, otherP) then
                            dist = R_PointToDist2(playerMo.x, playerMo.y, otherP.mo.x, otherP.mo.y)
                            if dist <= targetingDistThresScaled then
                                angle = -abs(R_PointToAngle2(
                                    playerMo.x,
                                    playerMo.y,
                                    otherP.mo.x,
                                    otherP.mo.y
                                ) - playerMo.angle)
                                if angle > closestAngle then
                                    target = otherP
                                    closestAngle = angle
                                end
                            end
                        end
                    end
                    break
                --never target players behind
                elseif angle > TARGETING_ANY_ANGLE_THRES and dist < closestDist then
                    target = otherP
                    closestDist = dist
                end
            end
        end
    end
    
    --print(target and ("target " .. target.name) or "no target")
    if target then
        local retiX = target.mo.x
        local retiY = target.mo.y
        local retiZ = target.mo.z + (target.mo.height >> 1)
        if player.dkrtargetreticle.frame and player.dkrtargetreticle.frame.valid then
            relocateTargetPiece(
                player.dkrtargetreticle.frame,
                retiX,
                retiY,
                retiZ,
                player.dkrtarget
            )
        else
            --spawn and position the reticle
            player.dkrtargetreticle.frame = P_SpawnMobj(retiX, retiY, retiZ, MT_DKRRETICLEFRAME)
            --allow it to trace back in case something weird happens
            player.dkrtargetreticle.frame.tracer = playerMo
        end
        if player.dkrtargetreticle.glass and player.dkrtargetreticle.glass.valid then
            relocateTargetPiece(
                player.dkrtargetreticle.glass,
                retiX,
                retiY,
                retiZ,
                player.dkrtarget
            )
        else
            --spawn and position the reticle
            player.dkrtargetreticle.glass = P_SpawnMobj(retiX, retiY, retiZ, MT_DKRRETICLEGLASS)
            --allow it to trace back in case something weird happens
            player.dkrtargetreticle.glass.tracer = playerMo
        end
        --update reticle color
        player.dkrtargetreticle.frame.color = player.skincolor
    else
        if player.dkrtargetreticle.frame and player.dkrtargetreticle.frame.valid then
            P_RemoveMobj(player.dkrtargetreticle.frame)
        end
        if player.dkrtargetreticle.glass and player.dkrtargetreticle.glass.valid then
            P_RemoveMobj(player.dkrtargetreticle.glass)
        end
        player.dkrtargetreticle = {}
    end
    player.dkrtarget = target
end

local function preuseDKRBackMagnet(p, cmd, ticsHeld, attackJustDown)
    refreshDKRTargetStuff(p)
    p.dkrmagnetblink = true
end

local function useDKRBackMagnet(p, cmd)
    local target = p.dkrtarget
    
    if canDKRTargetPlayer(p, target) then
        --you're going to brazil, and I'm brazil. C'mere
        table.insert(target.dkrmagnetisminfo, {
            timer = 0,
            timeUp = MAGNET_TIME,
            target = p,
            angRespect = p,
            electricity = {},
            soundTimer = 0,
            soundLoop = MAGNET_SOUND_LOOP
        })
		
		--hitfeed compat inserted by mine
		if (hitfeed) then
			local source = generateHFForPlayer(p)
			local victim = generateHFForPlayer(target)
			HF_SendHitMessage(source, victim, "H_DKRBM")
		end

        --fes compat also by mine
        if fesScoring then
            FES_scoring_rewardPlayerAction(p, "xitem_dkrmagnet")
        end
    end
    p.dkrtarget = nil
    if p.dkrtargetreticle.frame and p.dkrtargetreticle.frame.valid then
        P_RemoveMobj(p.dkrtargetreticle.frame)
    end
    if p.dkrtargetreticle.glass and p.dkrtargetreticle.glass.valid then
        P_RemoveMobj(p.dkrtargetreticle.glass)
    end
    p.dkrtargetreticle = {}
    p.kartstuff[k_itemamount] = $ - 1
    p.dkrmagnetblink = false
end

local function killMagnet(player, i)
    for j = 1, MAGNET_ELECTRICITY_COUNT do
        if player.dkrmagnetisminfo[i].electricity[j].valid then
            P_RemoveMobj(player.dkrmagnetisminfo[i].electricity[j])
        end
    end
    table.remove(player.dkrmagnetisminfo, i)
end

local function resetPlayerVars(player)
    player.dkrtarget = nil
    player.dkrtargetreticle = {}
    player.dkrmagnetisminfo = {}
    player.dkrmagnetblink = false
end

local function updateMagnetDrags(mobj)
    local player = mobj.player
    
    local didDrag = false
    local magnetMomX = 0
    local magnetMomY = 0
    
    if mustKillAllMagnets(player) then
        for i = #player.dkrmagnetisminfo, 1, -1 do
            killMagnet(player, i)
        end
        return
    end
    --EXIT EARLY
    
    --iterate backwards to be able to remove entries during iteration
    for i = #player.dkrmagnetisminfo, 1, -1 do
        local magnetInfo = player.dkrmagnetisminfo[i]
        local magnetTarget = magnetInfo.target
        
        if
            isMagnetTargetValid(magnetTarget)
            --timer check moved up here to prevent electricity being removed too early
            or magnetInfo.timer == magnetInfo.timeUp
            --never kill the magnet on the first tic b/c of distance so it at least does something
            or magnetInfo.timer
                and isMagnetAtInvalidRange(player, magnetTarget)
        then
            killMagnet(player, i)
        elseif canDKRBackMagnetPlayer(player) then
            local magnetMo = magnetTarget.mo
            --magnetangrespect should always be either the dragged or the target
            local magnetRespectMo = magnetInfo.angRespect.mo
            local destX, destY = magnetMo.x, magnetMo.y
            --aim for one side depending on which side is closer
            local relAng = R_PointToAngle2(mobj.x, mobj.y, destX, destY) - magnetRespectMo.angle
            --positive -> [0, 180) deg -> left of player -> right of target
            --negative -> [180, 360) deg -> right of player -> left of target
            local angOff
            
            if relAng < 0 then
                angOff = magnetRespectMo.angle + ANGLE_90
            else
                angOff = magnetRespectMo.angle - ANGLE_90
            end
            
            local sideOff = mobj.radius + magnetTargetOffsetDistScaled + magnetMo.radius
            
            destX = $ + FixedMul(sideOff, cos(angOff))
            destY = $ + FixedMul(sideOff, sin(angOff))
            
            local finalAng = R_PointToAngle2(mobj.x, mobj.y, destX, destY)
            
            didDrag = true
            magnetMomX = $ + P_ReturnThrustX(nil, finalAng, magnetPullForceScaled)
            magnetMomY = $ + P_ReturnThrustY(nil, finalAng, magnetPullForceScaled)
            
            --electricity effect
            for j = 1, MAGNET_ELECTRICITY_COUNT do
                local elec = magnetInfo.electricity[j]
                if not (elec and elec.valid) then
                    magnetInfo.electricity[j] = P_SpawnMobj(0, 0, 0, MT_DKRMAGNETELECTRICITY)
                    elec = magnetInfo.electricity[j]
                end
                elec.angle = randomInt()
                if P_RandomKey(2) then
                    elec.eflags = $ | MFE_VERTICALFLIP
                else
                    elec.eflags = $ & ~MFE_VERTICALFLIP
                end
            end
            
            if magnetInfo.soundTimer == magnetInfo.soundLoop then
                --kill a sound and replay it
                S_StopSoundByID(mobj, sfx_atdrmg)
                S_StartSound(mobj, sfx_atdrmg)
                --1, not 0, to account for being == to loop is technically 0
                magnetInfo.soundTimer = 1
            else
                if not magnetInfo.soundTimer then
                    --only play a new sound
                    S_StartSound(mobj, sfx_atdrmg)
                end
                magnetInfo.soundTimer = $ + 1
            end
            
            magnetInfo.timer = $ + 1
        end
    end
    
    if didDrag then
        mobj.momx = magnetMomX
        mobj.momy = magnetMomY
        --for some reason, if speed is 0, racers don't slide after momentum is set
        if player.speed <= 0 and P_IsObjectOnGround(mobj) then
            mobj.z = ternaryOperation(mobj.eflags & MFE_VERTICALFLIP, mobj.z - 1, mobj.z + 1)
        end
    end
end

local function shouldStopTargeting(player)
    return player.playerstate ~= PST_LIVE
        or player.exiting
        or player.kartstuff[k_spinouttimer]
        or player.kartstuff[k_squishedtimer]
        or player.kartstuff[k_respawn]
end

local function doPlayerMobjThinkerStuff(mobj)
    if leveltime == 0 and not defrostingDebounce or mobj.player.spectator then
        resetPlayerVars(mobj.player)
    else
        if not mobj.player.dkrmagnetisminfo then
            resetPlayerVars(mobj.player)
        end
        
        if shouldStopTargeting(mobj.player) then
            local player = mobj.player
            player.dkrtarget = nil
            player.dkrtargetreticle = {}
            player.dkrmagnetblink = false
        end
            
        updateMagnetDrags(mobj)
    end
end

addHook("MobjThinker", doPlayerMobjThinkerStuff, MT_PLAYER)

local function doDKRReticleThinkerStuff(mobj)
    local playerMo = mobj.tracer
    if not (playerMo and playerMo.valid) then
        return
    end
    
    local player = playerMo.player
    if not (player and player.valid)
        or not player.dkrtarget
        or not player.dkrtargetreticle.frame
        or not player.dkrtargetreticle.glass
        or player.spectator
        or shouldStopTargeting(player)
    then
        P_RemoveMobj(mobj)
    end
end

addHook("MobjThinker", doDKRReticleThinkerStuff, MT_DKRRETICLEFRAME)
addHook("MobjThinker", doDKRReticleThinkerStuff, MT_DKRRETICLEGLASS)

local function updateMapObjectScaledVars()
    targetingDistThresScaled = getMapObjectScaled(TARGETING_DIST_THRES)
    magnetKillDistThresScaled = getMapObjectScaled(MAGNET_KILL_DIST_THRES)
    magnetTargetOffsetDistScaled = getMapObjectScaled(MAGNET_TARGET_OFFSET_DIST)
    magnetPullForceScaled = getMapObjectScaled(MAGNET_PULL_FORCE)
end

local function doThinkFrameStuff()
    if leveltime == 0
        if not defrostingDebounce then
            updateMapObjectScaledVars()
            defrostingDebounce = true
        end
    else
        if leveltime == 1 then
            defrostingDebounce = false
        end
        for player in players.iterate do
            --very confident if player.dkrmagnetisminfo is nonnil, player is valid for this 
            if player.dkrmagnetisminfo then
                for _, magnetInfo in ipairs(player.dkrmagnetisminfo) do
                    for _, elec in ipairs(magnetInfo.electricity) do
                        local playerMo = player.mo
                        P_SetOrigin(
                            elec,
                            playerMo.x,
                            playerMo.y,
                            ternaryOperation(
                                playerMo.eflags & MFE_VERTICALFLIP,
                                playerMo.z + playerMo.height - MAGNET_ELECTRICITY_OFFSET,
                                playerMo.z + MAGNET_ELECTRICITY_OFFSET
                            )
                        )
                    end
                end
            end
        end
    end
end

addHook("ThinkFrame", doThinkFrameStuff)

local function drawDKRBackMagnet(v, p, c)
    xItemLib.func.hudDrawItem(v, p, c, nil, p.dkrmagnetblink and 2 or nil)
end

if fesScoring then
	FES_scoring_addAction("xitem_dkrmagnet", "Magnet Accuracy", 25)
end

updateMapObjectScaledVars()
for player in players.iterate do
    resetPlayerVars(player)
end

xItemLib.func.addItem(
    "DKR_BACK_MAGNET",
    "DKR Self-Magnet",
    "ATIDKRBM",
    "ATIDRBMS",
    0,
    {0, 0, 0, 1, 4, 4, 2, 0, 0, 0},
    {2, 0},
    nil,
    useDKRBackMagnet,
    drawDKRBackMagnet,
    nil,
    nil,
    {0, {SPR_XITS, 4}},
    true,
    preuseDKRBackMagnet
)

xItemLib.func.setXItemModData("XITEM_DESCRIPTIONS", "DKR_BACK_MAGNET", {
	tipDescription1 = "Attracts racer in front of user",
	tipDescription2 = "Attraction harshly drags victim backwards",
	tipUsage1 = "Hold \130Fire\128 to lock on",
	tipUsage2 = "Release to use",
})